"""
Main Gateway class which manages all user modules.
"""

import contextlib
import json
import logging
import signal
import sys
import threading
import typing as t
from collections.abc import Iterable

from .events import Event, attach_handlers, merge_events
from .modules import init_library_module_events
from .UuAppClient import UuAppClient

UserModule = t.Any

logger = logging.getLogger(__name__)
loggerc = logging.getLogger(__name__ + ".Config")


class Gateway(contextlib.AbstractContextManager):
    """Main application entrypoint.

    Gateway is responsible for initializing all events (and all modules) by providing
    them configuration and app start/stop information.

    This class is a `ContextManager`. On enter, the :meth:`.start` method is called,
    on leave, the :meth:`.stop` method is called.

    Warning:
        When an exception is thrown inside the ``module_init`` or ``event_init`` function
        (usually calling class constructors), it is not supressed and leads to
        classical exception behaviour (ie. program termination).

        When an exception is thrown inside an event handler, the exception is typically
        catched and printed via logging. Consult corresponding event for more information.

    Examples:

        .. code-block:: python

            from uun_iot import Gateway:
            config_file="config.json"
            with Gateway(config_file) as g:
                while True:
                    g.stopev.wait()

        In this example, the :class:`Gateway` is initialized using the
        configuration file in a `with` block. Upon entering, the gateway is
        started. When an exeption occurs, the gateway is gracefully stopped
        using its :meth:`.stop` method -- this allows the `Gateway` to do
        cleanup and inform user modules (here are none) that they should exit,
        too.

    Configuration:

        The preffered styling of the JSON keys is ``camelCase``, ie. first letter
        lowercase, letters after spaces uppercase and finally, the spaces removed.

        The JSON configuration file has the following example structure:

        .. code-block:: json
            :force:

            {
                "gateway": {

                    "moduleTimers": {
                        "customModule1": {
                            "receive": 120,
                            "send": 400
                        },
                        "customModule2": 60
                    },

                    "moduleBackupStorage": {
                        "customModule1": "backup/customModule1.json",
                        "customModule2": {
                            "path": "backup/customModule2.json",
                            "limit": 50
                        }
                    },

                    "customModule2": {
                        "option1": "value1"
                    }
                },

                "uuApp": { ... },
                "uuThing": { ... },
                "oidcGrantToken": { ... }
            }

        On the other hand, a **minimal** configuration file is an empty JSON file:

        .. code-block:: json

            {}

        Meaningful keys and subkeys are:

            - ``gateway``: `optional`. Main configuration for this IoT application

                - ``moduleTimers``: `optional`. Core functionality, dictionary with
                  periods (in seconds) for ``@on("tick")`` events for corresponding
                  modules. The module IDs are keys, the periods are values.

                    - Multiple timers can be
                      specified by introducing a subobject with the timer ID and the
                      timer period. The timer ID corresponds to the event defined in
                      Python code as ``@on("tick", "timerId")``.

                - ``moduleBackupStorage``: `optional`. Applies to modules based on
                  :class:`~uun_iot.modules.Module.Module`. The format is module ID
                  as key and file path as value. This key is used to specify
                  location to which should unsent data from the module be saved,
                  see :class:`~uun_iot.modules.Module.Module` for more information.

                    - You can specify additional information, such that the storage
                      should be limited in size. For this, specify the size of the
                      storage in number of entries in ``limit`` and add the
                      original file path in the ``path`` key.

                - ``<moduleId>``: `optional`. A default place for the configuration specific to
                  the module with ID ``moduleId``. The structure is arbitrary and
                  depends on your needs.

            - keys for :class:`uun_iot.UuAppClient`. `Optional`. See documentation there for
              more information. If you want to use secure communication with uuApp,
              specify the details in keys

                - ``uuApp``
                - ``uuThing``
                - ``oidcGrantToken``


    Args:
        config_file: path to configuration JSON file. See :class:`Config` for
            information about format of the file.
        module_init: a function instantiating modules and returning the
            instances with registered handlers which are to be registered to events
            defined in this library
        event_init: a function with different signature for advanced use cases. Takes
            a config file path, config dictionary and UuAppClient and returns instantiated
            events with registered handlers (using :func:`attach_handlers`).
            Beware that if you want to use library modules (ConfifUpdater, SocketServer, Heartbeat),
            you need to incorporate :func:`uun_iot.modules.init_library_modules` into the function.
            It can be used to define custom
            events (as long as they derive :class:`Event`) or create event-module strong
            coupling if needed (see :class:`ConfigUpdater` or :class:`SocketServer`).
            For defining custom events, see documentation in :mod:`events` module.
    """

    _events: t.Iterable[Event]

    #: event set when the gateway is running
    runev: threading.Event
    stopev: threading.Event

    def __init__(
        self,
        config_file: t.Optional[str],
        module_init: t.Optional[
            t.Union[
                t.Callable[[t.Dict, UuAppClient], UserModule],
                t.Callable[[t.Dict, UuAppClient], t.Iterable[UserModule]],
            ]
        ],
        event_init: t.Optional[
            t.Callable[[t.Optional[str], t.Dict, UuAppClient], t.Iterable[Event]]
        ] = None,
    ):

        logger.debug("Starting gateway module system.")

        if not module_init and not event_init:
            raise ValueError("Both init methods cannot be empty")

        if module_init and event_init:
            raise ValueError("Specify only one init method")

        if config_file is None:
            config = {}
        else:
            try:
                with open(config_file, "r") as f:
                    config = json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError("Could not decode JSON config") from e
            except FileNotFoundError as e:
                raise ValueError(
                    "Provided config file '%s' does not exist", config_file
                ) from e

        if "gateway" not in config:
            config["gateway"] = {}

        uuapp_client = UuAppClient(config)

        if module_init:
            logger.debug("using basic module init")
            base_events = init_library_module_events(config_file, config, uuapp_client)
            user_modules = module_init(config, uuapp_client)
            if not isinstance(user_modules, Iterable):
                user_modules = [user_modules]
            user_events = attach_handlers(config, user_modules, base_events)
            # do not discard events needed for library modules even when user modules do
            # not need these events
            events = merge_events(base_events, user_events)
        elif event_init:
            logger.debug("using advanced event init")
            # trust that event_init properly calls and merges with library events
            # init_base_events
            # and registers handlers on its own using :func:`events.attach_handlers`
            events = event_init(config_file, config, uuapp_client)
        else:
            assert False

        logger.debug("Registered following events: %s", repr(events))
        if not events:
            raise ValueError("Nothing to do, no defined handlers/events.")
        self._events = events

        self.runev = threading.Event()
        self.stopev = threading.Event()
        self.stopev.set()
        logger.debug("module system initilized")

    def __enter__(self):
        """Call :meth:`.start`."""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """Call :meth:`.stop`."""
        self.stop()

    def start(self):
        """Start Gateway and all its associated events.

        Set self.runev, clear self.stopev, invoke all ``@on("start")`` start
        all timers  and wait for all of them to finish.

        This is no-op when the gateway is already started.
        """
        if not self.runev.is_set():
            for ev in self._events:
                ev.start()
            self.runev.set()
            self.stopev.clear()

    def stop(self):
        """Stop Gateway and all its associated events.

        Peacefully stop (wait to finish) all timers and associated
        ``@on("tick")`` handlers, invoke all ``@on("stop")`` handlers and wait
        for all of them to finish.

        This is no-op when the gateway is not running.
        """

        # this also stops on_start function (oneshots) polling for the event
        if self.runev.is_set():
            for ev in self._events:
                ev.stop()
            self.runev.clear()
            self.stopev.set()

    def signal_handler(
        self,
        sig: int,
        frame,
        additional_cleanup: t.Optional[t.Callable[["Gateway"], None]] = None,
    ):
        """Handler for predefined signals.

        The following signals are supported:

        =================   ===========
        functionality no.   description
        =================   ===========
        1                   :meth:`.stop` Gateway, ``del`` ete all modules, run ``additional_cleanup`` and exit
        =================   ===========

        =========== =============
        signal      functionality
        =========== =============
        ``SIGINT``  1
        ``SIGTERM`` 1
        ``SIGUSR1`` 1 and exit with error code ``1``
        ``SIGUSR2`` 1
        =========== =============

        The signals need to be explicitly registered with this method being the
        associated handler. Register signals as:

        .. code-block:: python

            import signal
            from uun_iot import Gateway
            with Gateway(...) as g:
                signal.signal(signal.SIGTERM, g.signal_handler)
                ...

        If you want to specify ``additional_cleanup``, define a partial
        function from this method with first two arguments left empty.

        Args:
            sig: caught signal
            frame: exception frame
            additional_cleanup: optional function to be run to cleanup the
                `Gateway`. The function takes the `Gateway` instance as an
                argument.
        """
        logger.debug("Received signal `%s`", signal.Signals(sig).name)
        if sig in [signal.SIGINT, signal.SIGTERM, signal.SIGUSR1, signal.SIGUSR2]:
            # systemd issues SIGTERM to stop a program
            # user defined signal SIGUSR1 -- is issued from some Module to stop gateway and exit
            self.stop()

            # explicitly delete an event before exiting to give it a chance to execute its __del__
            # and the same for user modules bound to events via handlers
            for ev in self._events:
                del ev

            if additional_cleanup is not None:
                additional_cleanup(self)

            ec = sig in [signal.SIGUSR1]
            # SIGUSR1 -- exit with error
            # SIGUSR2 -- exit without error
            sys.exit(int(ec))


# def old_config_on_tick(self):
#    """
#    Gets new configuration from the uuApp, validates it, restarts timers
#    (if needed based on the new configuration) and saves the new
#    configuration locally. It also notifies modules (via their
#    ``@on("update")``) about configuration update.
#    It is triggered by a tick event.
#    """
#
#    new_config = self._uucmd()
#
#    if not new_config:
#        # did not get a valid configuration
#        return
#
#    if new_config != self.config["gateway"]:
#
#        loggerc.debug("Received new configuration from server: %s", new_config)
#
#        # for loop: restart timers if needed
#        for id, mt in self.g._timers.items():
#            if id not in new_config["moduleTimers"]:
#                continue
#
#            # new timer values are presented
#            if (
#                self.config["gateway"]["moduleTimers"][id]
#                != new_config["moduleTimers"][id]
#            ):
#
#                # if id == "config":
#                #    # TODO: needs testing if utils.Timer correctly handles setting period without stopping first
#                #    mt.timer.period = new_config["moduleTimers"][id]
#                #    loggerc.info(f"experimentally updated `{id}`'s timer to {new_config['moduleTimers'][id]} s")
#                #    continue
#
#                if mt.multi:
#                    # multiple timers for one module
#                    if (
#                        self.config["gateway"]["moduleTimers"][id].keys()
#                        != new_config["moduleTimers"][id].keys()
#                    ):
#                        loggerc.error(
#                            "Update cannot introduce new timer_id for a module, only update timer values."
#                        )
#                        continue
#
#                    for timer_id, new_interval in new_config["moduleTimers"][
#                        id
#                    ].items():
#                        if (
#                            new_interval
#                            != self.config["gateway"]["moduleTimers"][id][timer_id]
#                        ):
#                            mt.timer[timer_id].stop()
#                            mt.timer[timer_id].period = new_config["moduleTimers"][id][
#                                timer_id
#                            ]
#                            mt.timer[timer_id].runonstart = False
#                            mt.timer[timer_id].start()
#                            loggerc.info(
#                                f"updated `{id}:{timer_id}` timer to {new_config['moduleTimers'][id][timer_id]} s"
#                            )
#                else:
#                    # single timer
#                    mt.timer.stop()
#                    mt.timer.period = new_config["moduleTimers"][id]
#                    mt.timer.runonstart = False
#                    mt.timer.start()
#                    loggerc.info(
#                        f"updated `{id}`'s timer to {new_config['moduleTimers'][id]} s"
#                    )
#
#        # update config, only gateway key, rest is unchanged
#
#        # PASSIVE CONFIGURATION UPDATE
#        # propagated passively to each module
#        # modules contain _config in form of config["gateway"] assigned at module's init
#        #   (other keys are not needed and only cause more typing)
#
#        # DOES NOT WORK:
#        # cannot do `self.config["gateway"] = new_config` as it would not change the config["gateway"] dict present in each module:
#        # self.config["gateway"]           --> { 'module_1' --> module_1_data_old, ... }
#        # self.config["gateway"] --(update)--> { 'module_1' --> module_1_data_new, ... }
#        # (some_module)._config            --> { 'module_1' --> module_1_data_old, ... } as
#        #   it was never reassigned inside the module
#
#        # DOES WORK:
#        # update every key(pointer) in config["gateway"] separately; module's _config still points to the same config["gateway"]
#        #   but every key inside config["gateway"] is correctly updated
#        # self.config["gateway"] --> { 'module_1'           --> module_1_data_old, ... }
#        # self.config["gateway"] --> { 'module_1' --(update)--> module_1_data_new, ... }
#        # (some_module)._config  --> { 'module_1'           --> module_1_data_new, ... } as
#        #   the module's config and self.config["gateway"] point to the same (updated object)
#
#        # new update cannot introduce new keys to ["gateway"] (which is to be expected -- cannot create objects at run-time without any defining code)
#        updated_values = False
#        for key in self.config["gateway"]:
#
#            if key not in new_config:
#                loggerc.warning(
#                    "Update does not contain a configuration key `%s`. Update is not deleting this key from current configuration but it will be deleted on next run of application.",
#                    key,
#                )
#                continue
#
#            if self.config["gateway"][key] != new_config[key]:
#                updated_values = True
#                loggerc.debug(
#                    "changed key `%s` from %s to %s",
#                    key,
#                    self.config["gateway"][key],
#                    new_config[key],
#                )
#                self.config["gateway"][key] = new_config[key]
#
#        if not updated_values:
#            loggerc.debug(
#                "New configuration introduced only new keys not present on old config, discarding update."
#            )
#            return
#
#        # allow module callbacks to alter the configuration before saving to disk
#        self._invoke_update_callback()
#
#        # save updated config to file in case of power outage/app restart
#        with open(self._config_file, "w") as f:
#            f.write(
#                json.dumps(self.config, indent=4)
#            )  # pretty print in case of manual editing
#
#        loggerc.info("Updated configuration.")



import logging
import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum, auto

from ..exceptions import InvalidModuleId
from ..typing import ModuleId
from ..utils import class_ids, instance_ids

logger = logging.getLogger(__name__)


EVENT_KEY = "_uuniot"


class EventType(IntEnum):
    START = auto()
    STOP = auto()
    CONF_UPDATE = auto()
    TICK = auto()
    EXTERNAL = auto()


# TODO: save _HandlerInfoInternal along with method
# turn it to HandlerInfo on extraction
# this would simplify the saving logic (as it does not have to use __set__name__ class
# handler)
@dataclass
class _HandlerInfoInternal:
    event_type: EventType
    args: t.Tuple
    kwargs: t.Dict


@dataclass
class HandlerInfo:
    event_type: EventType
    args: t.Tuple
    kwargs: t.Dict
    owner_id: ModuleId


F = t.TypeVar("F", bound=t.Callable)
FT = t.TypeVar("FT", bound=t.Callable)


# no idea why this works (with regard to typing) and the simpler function approach does not
class Register(t.Generic[F]):
    """
    Workaround till Python 3.12 which offers parametric functions Descriptor
    cannot be parametrized as it is outside of scope and passing t.Callable as
    argument of type Type[T] does not work in typecheckers
    """

    # @staticmethod
    # def dynamic_register(subscription, owner, name: str):
    #    """Dynamically register event handler.

    #        owner: owner class
    #        name: attribute name to save
    #    """
    #    subscription.stamp_function()
    #    setattr(owner, name, subscription)
    #    # explicitly call __set_name__ as it is not called automatically in this case
    #    getattr(owner, name).__set_name__(owner, name)

    @classmethod
    def register(cls, event_type: EventType, *args, **kwargs):
        """Decorator to register class methods as event handlers.

        The decorator marks the function as event handler by assigning an attribute
        with event details directly to the method.

        Events are supposed to provide correct signature of event handlers and use this internally.
        """

        class FunctionWrapper:
            """
            Used exclusively for dynamic handler registering and in order to stop
            side-effects. We need to assign event attribute to the handler and this
            approach does not modify attributes of the original.
            """

            def __init__(self, function_or_method):
                self.fn = function_or_method

            def __call__(self, *args, **kwargs):
                return self.fn(*args, **kwargs)

        class Descriptor(t.Generic[FT]):
            """
            Class-type decorator based on descriptors (__set_name__)
            """

            def __init__(self, fn: FT):
                self.fn = fn
                self.__doc__ = fn.__doc__

            # def link(self) -> None:
            #    """Used to dynamically register handlers.

            #    Warning:

            #        Dangerous side effects! Use with care. This method cannot be used to
            #        dynamically register only some instances of a class - the underlying
            #        technique directly modifies the class method and it changes it for
            #        ALL instances.

            #    Make the method into a handler.
            #    """
            #    self._add_event_attribute(self.fn)

            def extract(self) -> FT:
                """Returns the function with additional event handler data.

                Might be used to conditionally register a handler.

                Warning:
                    In order for this method to have no side effects (such as modifing
                    the original function), it needs to wrap the function in another
                    call layer. If that might be a problem (if performance of some very
                    frequently called function is in stake), do not use dynamic
                    registering or rather use the :meth:`.link` technique.

                Example:

                    This short example shows a (trivial) usage. It can be applied similarly to
                    any other event.

                    .. code-block: python

                        # this example shows
                        from uun_iot import EvStart
                        from uun_iot.events import Register
                        ENABLE_EVENT = True

                        class MyModule:
                            def __init__(self):
                                if ENABLE_EVENT:
                                    self.handler = EvStart.subscribe(self._handler).extract()

                            def _handler(self):
                                print("ticking")

                Args:
                    descriptor: the classical subscription without decorator, eg.
                        `EvTick.subscribe(my_handler)`
                """
                wrapper = FunctionWrapper(self.fn)
                self._add_event_attribute(wrapper)
                return wrapper  # type: ignore

            def _add_event_attribute(self, callable_obj) -> None:
                logger.debug(
                    "producing dynamic stand-alone handler for '%s' using function '%s'",
                    event_type,
                    callable_obj,
                )

                # include information in attribute of the decorated function
                # this works on function but not 'methods', i.e. bound functions. On
                # these, it throws error: 'method' object has no attribute '_uuniot'.
                # bound methods override getattr and setattr in order to lookup the
                # parameters in the wrapped function (.__func__).
                setattr(
                    (
                        callable_obj.__func__
                        if hasattr(callable_obj, "__func__")
                        else callable_obj
                    ),
                    EVENT_KEY,
                    _HandlerInfoInternal(event_type, args, kwargs),
                )

            def __set_name__(self, owner, name: str):
                """Called after assigning decorated method as an attribute to the class.

                Args:
                    owner: owner class
                    name: name of the attribute
                """
                # owner -- owner class of the decorated method
                _, owner_id = class_ids(owner)
                logger.debug(f"decorating {self.fn} of {owner_id}")

                if len(args) > 0:
                    normalized_args = args
                elif args == ():
                    normalized_args = []
                elif args is None:
                    normalized_args = []
                else:
                    normalized_args = list(args)
                doc_list_args = [
                    ",".join([f'"{arg}"' for arg in normalized_args]),
                    ",".join(
                        (f'{key}="{val}"' for key, val in kwargs.items())
                        if kwargs is not None
                        else []
                    ),
                ]
                doc_arguments = ",".join(filter(lambda x: x != "", doc_list_args))
                doc_arguments = f"({doc_arguments})" if doc_arguments != "" else ""
                self.fn.__doc__ = (
                    f"Registered to handle event ``@Ev{event_type.name}.subscribe{doc_arguments}.``\n\n"
                    + self.fn.__doc__
                    if self.fn.__doc__ is not None
                    else ""
                )

                setattr(
                    self.fn,
                    EVENT_KEY,
                    _HandlerInfoInternal(event_type, args, kwargs),
                )

                # then replace ourself (Descriptor) with the original method
                setattr(owner, name, self.fn)

            def __call__(self, *args, **kwargs):
                """
                Type checkers do not know descriptor protocol, hence this is not used during runtime used.

                (In case the method is assigned dynamically (e.g. in __init__),
                (__set_name__ is not called. Then, the Descriptor class itself is stored
                (as the method and it needs to have __call__ defined.
                """
                return self.fn(*args, **kwargs)

        return Descriptor[F]


# typing does not work for some reason, class properties are not supported well
# https://github.com/python/mypy/issues/2563
# class Register(t.Generic[F]):
#    """
#    Workaround till Python 3.12 which offers parametric functions Descriptor
#    cannot be parametrized as it is outside of scope and passing t.Callable as
#    argument of type Type[T] does not work in typecheckers
#    """
#
#    @classmethod
#    def register(cls, event_type: EventType, *args, **kwargs):
#        """Decorator to register class methods as event handlers.
#
#        The decorator marks the function as event handler by assigning an attribute
#        with event details directly to the method.
#
#        Events are supposed to provide correct signature of event handlers and use this internally.
#        """
#
#        def descriptor(fn: F) -> F:
#            logger.debug("decorating '%s'", fn)
#
#            # include information in attribute of the decorated function
#            setattr(
#                fn,
#                EVENT_KEY,
#                _HandlerInfoInternal(event_type, args, kwargs),
#            )
#            return fn
#
#        return descriptor


class Event(ABC):
    """
    Pass whole config along with "gateway" key.

    It provides a default implementation of ContextManager - i.e. it can be used in
    `with` statements. Default implementation is to call :meth:`.start` on enter and
    :meth:`.stop` on exit.
    """

    event_type: t.ClassVar[EventType]
    _handlers: t.Any
    _config: t.Optional[t.Dict]

    # @classmethod
    # @property
    # @abstractmethod
    # def event_type(cls) -> EventType:
    #    pass

    def __init__(
        self,
        config: t.Optional[t.Dict] = None,
        *,
        requires_config=False,
    ):

        if not hasattr(self, "event_type"):
            raise AttributeError("Event must define its `.event_type`!")

        if requires_config:
            if config is None:
                config = {"gateway": {}}
            if "gateway" not in config:
                raise ValueError(
                    "Pass full configuration to events, not only the gateway subdictionary."
                )
        self._config = config

    @classmethod
    def _register(cls, reg_t_: t.Optional[t.Type[Register[F]]] = None, *args, **kwargs):
        """Register a handler with signature and

        Use as `self._register(Register[typing.Callable[<signature>]],
        additional_data_to_be_stored_for_event_handler)` to enable handler
        signature type checking via :attr:`typing.Callable` signature. To
        typecheck for methods, you need to include first auxilary "self"
        argument in the signature.

        This saves handler information (event type) directly as an attribute to
        the method in :class:`HandlerInfo` structure along with specified
        `*args` and `**kwargs`.

        Args:
            reg_t_: Register class used for type-checking handler signatures
        """
        if reg_t_ is None:
            return Register[t.Callable].register(cls.event_type, *args, **kwargs)
        else:
            return reg_t_.register(cls.event_type, *args, **kwargs)

    @abstractmethod
    def add_handlers(
        self, handler_list: t.Iterable[t.Tuple[t.Callable, HandlerInfo]]
    ) -> None:
        pass

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc_info):
        self.stop()

    def number_of_handlers(self) -> int:
        return len(self._handlers)


def extract_event_handlers_from_object(
    obj, is_class=False
) -> t.Iterable[t.Tuple[t.Callable, HandlerInfo]]:
    """Find marked event handlers in an instance and output a dictionary with the handlers.

    Attributes starting with double underscore `__` are skipped.

    Args:
        obj: the instance in which to search for handlers
        is_class: whether the object is a class; ie. if it is instantiated

    Returns:
        an iterable with tuples of (function, handler_info)
    """
    handlers_and_info = []
    for atr in dir(obj):
        if atr.startswith("__"):
            continue
        fn = getattr(obj, atr, None)
        if callable(fn):
            event_info = getattr(fn, EVENT_KEY, None)
            if event_info is None:
                continue

            _, owner_id = class_ids(obj) if is_class else instance_ids(obj)
            if owner_id is None:
                raise InvalidModuleId(
                    f"Invalid or missing module id for module '{obj}'. The library tries to get the id from the '.id' attribute or it deduces it from class name. If you see this, please provide snakeCased module id in the '.id' attribute of the module."
                )
            handler_info = HandlerInfo(
                event_info.event_type, event_info.args, event_info.kwargs, owner_id
            )
            # it is problematic to delete from instance method
            # proper way would be to delete from a class
            # but that would prevent other instances to ever be registered
            # delattr(fn, EVENT_KEY)
            handlers_and_info.append((fn, handler_info))
    return handlers_and_info


# class Register(t.Generic[F]):
#    """
#    Workaround till Python 3.12 which offers parametric functions Descriptor
#    cannot be parametrized as it is outside of scope and passing t.Callable as
#    argument of type Type[T] does not work in typecheckers
#    """
#
#    @classmethod
#    def register(cls, event_type: EventType, *args, **kwargs):
#        """Decorator to register class methods as event handlers.
#
#        The decorator marks the function as event handler by assigning an attribute
#        with event details directly to the method.
#
#        Events are supposed to provide correct signature of event handlers and use this internally.
#        """
#
#        class Descriptor(t.Generic[FT]):
#            """
#            Class-type decorator based on __set_name__
#            """
#
#            def __init__(self, fn: FT):
#                self.fn = fn
#                self.__doc__ = fn.__doc__
#
#            def __set_name__(self, owner, name: str):
#                """Called after assigning decorated method as an attribute to the class.
#
#                Args:
#                    owner: owner class
#                    name: name of the attribute
#                """
#                # owner -- owner class of the decorated method
#                _, owner_id = class_ids(owner)
#                logger.debug(f"decorating {self.fn} of {owner_id}")
#
#                # include information in attribute of the decorated function
#                setattr(
#                    self.fn,
#                    EVENT_KEY,
#                    HandlerInfo(event_type, args, kwargs, owner_id),
#                )
#                # then replace ourself (Descriptor) with the original method
#                setattr(owner, name, self.fn)
#
#            def __call__(self, *args, **kwargs):
#                """In case the method is assigned dynamically (e.g. in __init__),
#                __set_name__ is not called. Then, the Descriptor class itself is stored
#                as the method and it needs to have __call__ defined.
#                """
#                return self.fn(*args, **kwargs)
#
#        return Descriptor[F]



class UunIotException(Exception):
    """Base UunIotException class"""


class UuAppClientException(UunIotException):
    """Internal error in request pre-processing occured."""


class TokenError(UuAppClientException):
    """Error when getting/validating a token occured."""


class TokenCommandError(TokenError):
    """Server error when getting a token occured."""


class EventException(UunIotException):
    """Exception in event management system."""


class UnsupportedEvent(EventException):
    """Unknown gateway event."""


class InvalidModuleId(UunIotException):
    """Module ids can be only strings."""


class EventCannotRegisterHandler(EventException):
    """An error occured while registering a new handler for event."""


# old:
class EventRegisterAlreadyInstantiated(EventException):
    pass


class EventRegisterNotYetInstantiated(EventException):
    pass



import itertools
import logging
import threading
import typing as t
from dataclasses import dataclass

from ..utils import classproperty
from .Event import Event, EventType, Register

logger = logging.getLogger(__name__)

SelfM = t.Any


class EvConfUpdate(Event):
    """
    Event fired when new gateway configuration is available.

    This event is fired using :class:`uun_iot.modules.ConfigUpdater.ConfigUpdater`
    """

    event_type = EventType.CONF_UPDATE
    _handlers: t.List[t.Callable]

    def __init__(self, config=None):
        self._handlers = []
        super().__init__(config=None)

    @classproperty
    def subscribe(cls):
        """Register config update event.

        Register a handler to obtain the new configuration - it is passed in a single
        positional argument to the handler.

        If you are using the utility class Module, it is already registered for you. If
        you plan to register your own handler for some reason, consult documentation in
        Module.
        """
        return cls._register(Register[t.Callable[[SelfM, t.Dict], None]])

    def add_handlers(self, handler_list):
        self._handlers += [fn for fn, info in handler_list]

    def start(self):
        """Noop."""
        logger.debug(
            "EvConfUpdate has these handlers registered: %s", repr(self._handlers)
        )

    def stop(self):
        """Noop"""

    def fire_handlers(self, new_config: dict, block: bool = True):
        """Execute all stored handlers with new configuration.

        They are each executed in new thread.

        Args:
            new_config ([TODO:type]): [TODO:description]
            block: block till all of the executed handlers finish. Default is True.
        """
        threads = []
        for fn in self._handlers:
            logger.debug("Calling EvConfUpdate handler, function `%s`", fn)
            t = threading.Thread(target=fn, args=(new_config,))
            t.start()
            threads.append(t)

        if block:
            for t in threads:
                t.join()



import logging
import threading
import typing as t
from dataclasses import dataclass

from ..exceptions import EventCannotRegisterHandler
from ..typing import ModuleId
from ..utils import classproperty
from .Event import Event, EventType, Register

logger = logging.getLogger(__name__)

SelfM = t.Any
HandlerStartT = t.Callable[[SelfM, t.Tuple[threading.Event, threading.Event]], None]


class EvStart(Event):
    """Event for registering handlers to run on start of app.

    It is possible to register both blocking and non-blocking handlers.

    Warning:

        Ordering of handler execution is not given and cannot be relied upon.

    Examples:

        - configuration:

            .. code-block:: json

                {}

        .. code-block:: python

            from uun_iot import EvStart
            class AdvancedDataMeasurement(Module):
                @EvStart.subscribe
                def oneshot(self, evs):
                    # polling for data
                    runev, stopev = evs
                    while runev.is_set():
                        print("Polling for voltage reading from voltmeter...")
                        data = 53.8
                        if data > 50:
                            time.sleep(1)
                        else:
                            time.sleep(1.5)
                            print(data)

    """

    event_type = EventType.START
    _handlers: t.List[t.Tuple[t.Callable, bool]]
    _threads: t.Iterable[t.Tuple[threading.Thread, bool]]
    runev: threading.Event
    stopev: threading.Event

    def __init__(self, config=None):
        self.runev = threading.Event()
        self.stopev = threading.Event()
        self.stopev.set()
        self._handlers = []
        super().__init__(config=None)

    @classproperty
    def subscribe(cls):
        """Subscribe to Start event.

        The handler is run in a separate thread. The `.stop` method blocks until the
        thread returns.
        """
        return cls._register(Register[HandlerStartT], block=False)

    @classproperty
    def subscribe_blocking(cls):
        """This is a blocking handler variant.

        The handler is run in a separate thread and the `.start` method waits until the
        handler returns.
        """
        return cls._register(Register[HandlerStartT], block=True)

    def add_handlers(self, handler_list):
        self._handlers += [(fn, info.kwargs["block"]) for fn, info in handler_list]

    def start(self):
        """Invoke handlers in separate threads.

        Blocks until those registered in `@EvStart.subscribe_blocking` finish.
        """
        logger.debug("EvStart has these handlers registered: %s", repr(self._handlers))
        if not self._handlers or self.runev.is_set():
            # if no handlers to run or was not stopped
            return

        self.runev.set()
        self.stopev.clear()

        threads = []
        for fn, block in self._handlers:
            # do not store the thread to kill it later, rely solely on the func
            #   to terminate itself when runev Event is unset
            logger.debug(
                "Calling %sEvStart handler, function `%s`",
                "(blocking) " if block else "",
                fn,
            )
            th = threading.Thread(
                target=fn, args=((self.runev, self.stopev),), name=repr(fn)
            )
            th.start()
            threads.append((th, block))

        for th, block in threads:
            if block:
                logger.info(
                    "Waiting for blocking handler thread '%s' to finish in EvStart.start",
                    th.name,
                )
                th.join()

        self._threads = threads

    def stop(self):
        """Block till handlers terminate.

        Set `runev`, unset `stopev` and wait for started threads to finish.
        """
        # do not kill thread, rely solely on the handler
        #   to terminate itself when runev Event is unset
        if not self.runev.is_set():
            return

        self.runev.clear()
        self.stopev.set()

        for th, _ in self._threads:
            logger.debug("stopping %s", th)
            th.join()
            logger.debug("stopped %s", th)

        self._threads = []



import contextlib
import logging
import threading
import typing as t
from collections import defaultdict

from ..exceptions import EventCannotRegisterHandler
from ..typing import ModuleId
from ..utils import RepeatTimer, classproperty
from .Event import Event, EventType, HandlerInfo, Register

logger = logging.getLogger(__name__)
SelfM = t.Any

CONFIG_TIMER_KEY = "moduleTimers"
GATEWAY_KEY = "gateway"

HandlerTickT = t.Callable[[SelfM], None]


class EvTick(Event, contextlib.AbstractContextManager):
    """
    Event that fires periodically when a timer expires.

    Periods are loaded from configuration file and are updated on configuration update.

    Pass full configuration.

    Examples:

        - ``timer`` event without ID

            - configuration:

                .. code-block:: json

                    {
                        "gateway": {
                            "moduleTimers": {
                                "timerModule": 1
                            }
                        }
                    }

            .. code-block:: python


                from uun_iot import EvTick
                class TimerModule:
                    id = "timerModule" # optional
                    @EvTick.subscribe
                    def periodical(self):
                        print("Tick tock every 1 s.")

        - ``timer`` event with ID

            - configuration

                .. code-block:: json

                    {
                        "gateway": {
                            "moduleTimers": {
                                "sendReceive": {
                                    "send": 2,
                                    "get": 1
                                }
                            }
                        }
                    }

            .. code-block:: python

                class SendReceive:
                    @EvTick.subscribe_timer("get")
                    def retrieve(self):
                        print(f"Retrieving data...")

                    @EvTick.subscribe_timer("send")
                    def send(self):
                        print(f"Sending data...")
    """

    event_type = EventType.TICK

    _fire_on_start: bool

    startev: threading.Event
    _handlers: t.DefaultDict[
        ModuleId, t.DefaultDict[t.Optional[str], t.List[t.Callable]]
    ]
    _timers: t.Collection[t.Tuple[RepeatTimer, t.Tuple[ModuleId, t.Optional[str]]]]

    def __init__(self, config=None, fire_on_start: bool = True):
        """
        Args:
            config: dictionary with configuration
            fire_on_start: if the timers should execute right when :meth:`.start` is run
                if True, run on start and then each period,
                if False, do not run on start but after that run each period
        """
        self.startev = threading.Event()
        self._handlers = defaultdict(lambda: defaultdict(list))  # handlers[x][y]=[...]
        self._timers = []
        self._fire_on_start = fire_on_start
        super().__init__(config=config, requires_config=True)

    @classproperty
    def subscribe(cls):
        """Decorate a method to register it as a EvTick handler

        The handler corresponds to the timer found in config under
            ``"moduleTimers: { "moduleId": <period in seconds [float]>}``
        """
        return cls._register(Register[HandlerTickT], None)

    @classmethod
    def subscribe_timer(cls, timer_id: str):
        """Decorate a method to register it as a EvTick handler

        The handler corresponds to the timer found in config under
            ``"moduleTimers: { "moduleId": { "timerId": <period in seconds [float]>}}``

        Args:
            timer_id: timer id corresponding to configuration entry and current class
        """
        if not isinstance(timer_id, str):
            raise EventCannotRegisterHandler("Only string timer IDs are supported.")
        return cls._register(Register[HandlerTickT], timer_id)

    def add_handlers(
        self,
        handler_list: t.Iterable[t.Tuple[t.Callable, HandlerInfo]],
    ):
        if self.startev.is_set():
            raise EventCannotRegisterHandler(
                "Stop the EvTick before adding new handlers."
            )

        handlers = self._handlers
        # populate and check validity of handlers
        for fn, info in handler_list:
            mid = info.owner_id
            timer_id = info.args[0]
            handlers[mid][timer_id].append(fn)
            if None in handlers[mid] and len(set(handlers[mid].keys())) > 1:
                # multiple None handlers are valid - different instances
                # of same class might register None handler each
                raise EventCannotRegisterHandler(
                    "You cannot mix @EvTick.handler and @EvTick.handler_timer"
                    f" in module '{mid}'"
                )

        assert self._handlers == handlers
        # default dict no longer needed for reading, convert to dict to prevent accidental bugs
        self._timers = self._init_timers(
            dict(handlers),
        )

    def _init_timers(
        self,
        handlers: t.Dict[ModuleId, t.Dict[t.Optional[str], t.List[t.Callable]]],
    ):
        """Initialize timers for handlers from configuration."""
        assert self._config is not None
        timers = []
        for mid in handlers:
            try:
                interval_base = self._config[GATEWAY_KEY][CONFIG_TIMER_KEY][mid]
                assert interval_base != {}
            except (KeyError, AssertionError):
                # user has registered tick handler but no timer entry in config
                logger.warning(
                    "You defined handler for EvTick but did not specify"
                    f" timer period for module '%s'.",
                    mid,
                )
                continue

            interval_dict: t.Dict[t.Optional[str], float]
            if not isinstance(interval_base, dict):
                interval_dict = {None: interval_base}
            else:
                interval_dict = interval_base

            set_handlers = set(handlers[mid].keys())
            set_config = set(interval_dict.keys())
            if set_handlers != set_config:
                logger.error(
                    "Mismatch between defined EvTick handlers and timers in"
                    " configuration '.moduleTimers'."
                    " In configuration, there are '%s' and you defined handlers '%s'.",
                    set_config,
                    set_handlers,
                )
                raise EventCannotRegisterHandler()
            for timer_id, fns in handlers[mid].items():
                period = interval_dict[timer_id]
                timers.extend(
                    (
                        RepeatTimer(period, fn, runonstart=self._fire_on_start),
                        (mid, timer_id),
                    )
                    for fn in fns
                )
        return timers

    def number_of_handlers(self):
        """Get number of registered EvTick handlers.

        Warning: See :meth:`.number_of_timers` for an actual number of functional timers.
        """
        s = 0
        for mid, subev in self._handlers.items():
            for timer_id, fns in subev.items():
                s += len(fns)
        return s

    def number_of_timers(self):
        """Get number of created timers, as specified in configuration.

        Note that :meth:`.number_of_handlers` always returns number greater than or equal to
        :meth:`.number_of_timers`

        Some handlers might not have a timer entry in configuration and the
        corresponding timer is thus not created. Still, the handler is stored and a
        future configuration update might add the timer for the handler.
        """
        return len(self._timers)

    def start(self):
        logger.debug("EvTick has these handlers registered: %s", repr(self._handlers))
        logger.debug("EvTick has these timers registered: %s", repr(self._timers))
        for timer, (mid, tid) in self._timers:
            timer.start()
            logger.debug("Timer `%s:%s` started.", mid, tid)

    def stop(self):
        for timer, (mid, tid) in self._timers:
            timer.stop()
            logger.debug("Timer `%s:%s` stopped.", mid, tid)

    def update(self, new_config: dict):
        """Update timers from config. Compare new periods to existing and
        restart timer if neccesarry.

        The possible timer restart stops the timer, sets new period and then run the
        timer with new period, skipping the initial fire on start.

        Side effect: should the EvTick event be stopped and started again,
        updated timers will have runonstart=False.
        """
        self._config = new_config

        for timer, (mid, tid) in self._timers:
            try:
                new_period = self._config[GATEWAY_KEY][CONFIG_TIMER_KEY][mid]
                if tid is not None:
                    new_period = self._config[GATEWAY_KEY][CONFIG_TIMER_KEY][mid][tid]
            except KeyError:
                logging.warning(
                    "Configuration update removed timer entry for timer '%s.%s'.",
                    mid,
                    tid,
                )
                continue

            if not isinstance(new_period, (int, float)):
                logger.warning(
                    "Received invalid period in new configuration %s.%s: '%s'",
                    mid,
                    tid,
                    timer,
                )
                continue

            if new_period != timer.period:
                timer.stop()
                timer.runonstart = False
                timer.period = new_period
                timer.start()
                logger.info("Updated `%s.%s`'s timer to {new_period} s", mid, tid)



import copy
import logging
import typing as t

from ..events import EvTick
from ..typing import ModuleId

logger = logging.getLogger(__name__)


class ConfigUpdater:
    """Configuration update module.

    Gets new configuration from the uuApp, validates it, calls supplied callbacks on config update, restarts timers
    (if needed based on the new configuration) and saves the new
    configuration locally. It also notifies modules (via their
    ``@EvUpdate``) about configuration update.
    It is triggered by a tick event.

    update config, only gateway key, rest is unchanged

    PASSIVE CONFIGURATION UPDATE
        propagated passively to each module
        modules contain _config in form of config["gateway"] assigned at module's init
          (other keys are not needed and only cause more typing)

    DOES NOT WORK:
        cannot do `self.config["gateway"] = new_config` as it would not change the gconfig=config["gateway"] dict present in each module:

        >>> config = {"gateway": {"key": 1}}
        >>> sub = config["gateway"]
        >>> id(config)
        136617035769600
        >>> id(config["gateway"])
        136617037635904
        >>> id(sub)
        136617037635904
        >>> config["gateway"] = {"new": 2} # new dictionary {new: 2} is created in memory and reference saved in config["gateway"]
        >>> id(config)
        136617035769600
        >>> id(config["gateway"])
        136617035590272
        >>> id(sub)
        136617037635904
        >>> sub
        {'key': 1}

        self.config["gateway"]           --> { 'module_1' --> module_1_data_old, ... }
        self.config["gateway"] --(update)--> { 'module_1' --> module_1_data_new, ... }
        (some_module)._config            --> { 'module_1' --> module_1_data_old, ... } as
          it was never reassigned inside the module

    DOES WORK:
        update every key(pointer) in config["gateway"] separately; module's _config still points to the same config["gateway"]
          but every key inside config["gateway"] is correctly updated
        self.config["gateway"] --> { 'module_1'           --> module_1_data_old, ... }
        self.config["gateway"] --> { 'module_1' --(update)--> module_1_data_new, ... }
        (some_module)._config  --> { 'module_1'           --> module_1_data_new, ... } as
          the module's config and self.config["gateway"] point to the same (updated object)

    Args:
        config: IMPORTANT: pass whole configuration, not just `gateway` key.
            This module saves new configuration to file.
        uucmd_update: [TODO:description]
        update_config_callback: new configuration (only the "gateway" subkey) will be
            passed to this function.
    """

    id: ModuleId = "config"

    def __init__(
        self,
        config: t.Dict,
        uucmd_update: t.Callable[[], t.Optional[t.Dict]],
        save_config: t.Callable[[t.Dict], None],
        update_config_callback: t.Callable[[t.Dict], None],
    ):
        if "gateway" not in config:
            raise ValueError(
                "ConfigUpdater operates on the whole configuration, do not pass only the 'gateway' key"
            )
        self._config = config
        self._save_config = save_config
        self._uucmd = uucmd_update
        self._update_callback = update_config_callback

    @EvTick.subscribe
    def update_config(self):
        fetched_config = self._uucmd()

        # if True, do not discard now non-existent keys in new config
        nondecremental_update = False
        if "nondecrementalUpdate" in self._config["gateway"] and self._config["gateway"]["nondecrementalUpdate"]:
            nondecremental_update = True

        # avoid mutating reference in existing configs
        dup_config = copy.deepcopy(self._config)
        if not fetched_config or fetched_config == dup_config["gateway"]:
            return

        if nondecremental_update:
            gconf: dict = dup_config["gateway"]
            gconf.update(fetched_config)
            dup_config["gateway"] = gconf
        else:
            dup_config["gateway"] = fetched_config

        #for key in dup_config["gateway"]:
        #    if key not in fetched_config:
        #        logger.warning(
        #            "Update does not contain a configuration key `%s`."
        #            " Update is not deleting this key from current configuration"
        #            " but it will be deleted on next run of application.",
        #            key,
        #        )
        #        continue

        #    if dup_config["gateway"][key] != fetched_config[key]:
        #        logger.debug(
        #            "changed key `%s` from %s to %s",
        #            key,
        #            dup_config["gateway"][key],
        #            fetched_config[key],
        #        )
        #        dup_config["gateway"][key] = fetched_config[key]

        # self._config["gateway"] = new_config # passive update
        # self._config = {"gateway": new_config} # decouple
        self._config = dup_config
        self._save_config(self._config)
        self._update_callback(self._config["gateway"])
